/*
File: ServerProtocol.java
Date		Author		Changes
Aug 24 96	Lukasz Beca	Created
Aug 26 96       Lukasz Beca     Messages JASK and JANS added
Nov 15 96       Lukasz Beca     Bug in handle*ANS eliminated
Dec 02 96       LUkasz Beca     wake() eliminated from all functions
*/

package webwisdom.tango.protocols;

import java.net.*;
import java.util.*;
import webwisdom.tango.protocols.Protocol;
import webwisdom.tango.messages.*;
import webwisdom.tango.threads.*;
import webwisdom.tango.consts.Const;
import webwisdom.tango.structures.*;

public class ServerProtocol extends Protocol{


  static private int appno = 1;
  static private int SIDno = 1;
  protected int UID;
  private boolean logged_in;

  protected ThreadOut out;
  protected ComThread c;
  protected DbThread d;
  protected Timeout _t;  

 public ServerProtocol(ThreadOut outparam, ComThread cparam, DbThread dparam, Timeout t) {

  out = outparam;
  c = cparam;
  d = dparam;
  _t = t;
  UID = 0;
  logged_in = false;
    
 }	

 public void handleJOIN(Message a) {


  JoinMessage m = (JoinMessage) a;
  int SID = m.getSID();
  int AT = m.getAT();
  int uniqueAID; 
  CommandMessage comm;
  //SupdMessage update, mupdate;
  ThreadOut tmp;
  int masterUID;
  int newSID;
  //String name;
  JaskMessage jask;


  System.out.println("T " + UID + " ServerProtocol JOIN: JOIN message accepted");
  System.out.println("T " + UID + " ServerProtocol JOIN: Data AT " + AT + " SID " + SID);
 
   if (!logged_in) {
       System.out.println("T " + UID + " ServerProtocol JOIN: User not logged in, operation not permited");
       
       return;

    }

  //if (AI in SID for UID(UID)) stop

  synchronized(d) {    // start of operations on system state data
      
  if (!d.getApplications().inSession(SID, UID)) {

   

    //if SID == 0

    if(SID == 0) {

      // SID  = generate new SID    
      newSID = assignSID();
     
      //INSERT INTO SESSIONS (SID, AT, MASTER)
      d.getSessions().addSession(newSID, AT, UID);
      c.getSessionTable().addSession(newSID);
         
      System.out.println("T " + UID  + " ServerProtocol JOIN: new session created, SID " + newSID);

    //INSERT INTO APPLICATIONS (AID, SID, UID(UID))

      System.out.println("T " + UID  + " ServerProtocol JOIN: type of new session " + d.getSessions().getType(newSID));
	
     
	//generate unique AID
      uniqueAID = assignAID();

      d.getApplications().addApplication(uniqueAID, newSID, UID);
      //   c.getSessionTable().addApplication(newSID, out, uniqueAID);
    
      System.out.println("T " + UID  + " ServerProtocol JOIN: application " + uniqueAID + " added to session " + newSID + " and disabled");

      //SEND COMMAND(AT, params, AID, newSID)

      comm = new CommandMessage(AT, "EMPTY", uniqueAID, newSID);
      
      out.putMessage(comm);             
     
    }
    else {

      //SELECT mUID = master FROM SESSIONS WHERE SID = SID
      // SEND_TO (mUID, JASK(UID, SID))

      masterUID = d.getSessions().getMaster(SID);

      if (masterUID == -1) {
	System.out.println("Server Protocol JOIN: session " + SID + " does not exist, operation not permitted");
	return;

      }
      tmp = c.getThread(masterUID);
      jask = new JaskMessage(AT, SID, UID);
      tmp.putMessage(jask);
            
    } 

  }
  else {
     System.out.println("Server Protocol JOIN: user already in session");
  }
  
  }  //end of operations on system state data

  d.getApplications().list();
  d.getSessions().list();

}

 public void handleJANS(Message a) {

  JansMessage m = (JansMessage) a;
  int AT = m.getAT();
  int SID = m.getSID();
  int lUID = m.getUID();
  int answer = m.getAnswer();
  int uniqueAID;
  ThreadOut tmp;
  //SupdMessage update;
  CommandMessage comm;
  //String name;


  System.out.println("T " + UID  + " ServerProtocol JANS: JANS message accepted");
  System.out.println("T " + UID  + " ServerProtocol JANS: Data AT " + AT + " SID " + SID + " UID " + lUID + " answer " + answer);
 
   if (!logged_in) {
       System.out.println("T " + UID  + " ServerProtocol JANS: User not logged in, operation not permited");
       
       return;

    }     

  synchronized(d) {   //start of operations on system state data

   if (!d.getParticipants().isLoggedIn(lUID)) {

      System.out.println("Server Protocol JANS: user " + lUID + " not logged in");
     return;
   }

   //IF "no" stop

   if (answer == Const.NO) {

     System.out.println("T " + UID  + " ServerProtocol JANS: Master does not agree for joining session");
     
     tmp = c.getThread(lUID); 
     if (tmp != null) {
       tmp.putMessage(a);
     }
     else {
      System.out.println("T " + UID  + " ServerProtocol JANS: user " + lUID + " not logged in");
     } 
     return;
   }

  
   if (d.getSessions().getMaster(SID) == -1) {

      System.out.println("T " + UID  + " ServerProtocol JANS: session " + SID + " does not exist");
      return;
   }



   if (d.getApplications().inSession(SID, lUID)) {

     System.out.println("T " + UID  + " ServerProtocol JANS: user " + lUID + " already in session " + SID);
     return;
   }


   
   if (AT != d.getSessions().getType(SID)) { // check whether type is correct

     System.out.println("Server Protocol JANS: type of new application is not correct ");
     return;
   }

     //generate unique AID
   uniqueAID = assignAID();
   tmp = c.getThread(lUID);

     //INSERT INTO APPLICATIONS(AID, SID, lUID)
   d.getApplications().addApplication(uniqueAID, SID, lUID);
   //c.getSessionTable().addApplication(SID, tmp, uniqueAID);
    
   System.out.println("Server Protocol JANS: application " + uniqueAID + " added to session " + SID + " and disabled");

      //SEND_TO(UID, COMMAND(AT, params, AID, SID))

   

   comm = new CommandMessage(AT, "EMPTY", uniqueAID, SID);
      
   tmp.putMessage(comm);

  //  name = d.getUsers().getName(lUID);

//    update = new SupdMessage(Const.ADDUPDATE, SID, AT, lUID, name);    

//       SEND_TO_ALL SUPD(add, SID, UID(UID))

//    sendToAll(update);
     
   } //end of operations on system state data   
  

}



  public void handleSTARTED(Message a) {

    StartedMessage m = (StartedMessage) a;
    int AID = m.getAID();
    int messageSID = m.getSID();
    int SID;
    SupdMessage update, mupdate;
    int type;
    String name;

    System.out.println("T " + UID  + " ServerProtocol STARTED: STARTED message accepted");
    System.out.println("T " + UID  + " ServerProtocol STARTED: Data AID: " + AID);

    if (!logged_in) {
     System.out.println("T " + UID  + " ServerProtocol STARTED: User not logged in, operation not permited");
     return;

    }
    
    
    synchronized(d) {   //start of operations on system state data
    
    SID = d.getApplications().getSession(AID);

    if (SID == -1) {

     System.out.println("T " + UID  + " ServerProtocol STARTED: Application " + AID + " does not exists, cannot enable");
     update = new SupdMessage(Const.REMOVEUPDATE, messageSID, 0, 0, "");  //AT = 0;
     out.putMessage(update);
     return;

    }

    d.getApplications().enableApp(AID);
    
    if (!d.getSessions().isEnabled(SID)) {
    
      d.getSessions().enableSession(SID);
    
    }
    c.getSessionTable().addApplication(SID, out, AID);
    System.out.println("T " + UID  + " ServerProtocol STARTED: Application " + AID + " enabled");

    name = d.getUsers().getName(UID);
    type = d.getSessions().getType(SID);

    if (d.getSessions().getMaster(SID) == UID) {

      System.out.println("T " + UID  + " ServerProtocol STARTED: Application " + AID + " is master");

      update = new SupdMessage(Const.ADDUPDATE, SID, type, UID, name);
         
      mupdate =  new SupdMessage(Const.MASTERUPDATE, SID, type, UID, name);
     
      //SEND_TO_ALL SUPD(add, SID, AT, UID(UID), username)

      sendToAll(update);

      // SEND_TO_ALL SUPD(master, SID, 0, UID(UID), username);

      sendToAll(mupdate);
      

    } else {

      System.out.println("T " + UID  + " ServerProtocol STARTED: Application " + AID + " is not master");

      update = new SupdMessage(Const.ADDUPDATE, SID, type, UID, name);    

      //SEND_TO_ALL SUPD(add, SID, UID(UID))

      sendToAll(update);

    }
    
    
    } // end of operations on system state data

  }

 public void handleFINISH(Message a) {

  FinishMessage m = (FinishMessage) a;
  int AID = m.getAID();
  int user;

  //SELECT s = SID FROM APPLICATIONS WHERE aid = AID
  //DELETE FROM APPLICATIONS WHERE aid = AID
  //SELECT M = S.master FROM SESSIONS WHERE aid = AID
  //IF (M == UID(UID)) DELETE FROM APPLICATIONS WHERE SID = s
  
  System.out.println("T " + UID  + " ServerProtocol FINISH: FINISH message accepted");
  System.out.println("T " + UID  + " ServerProtocol FINISH: Data AID: " + AID);


  if (!logged_in) {
     System.out.println("T " + UID  + " ServerProtocol FINISH: User not logged in, operation not permited");
     return;

  }
  
  synchronized(d) {   //start of operations on system state data
   
  user = d.getApplications().getUser(AID);

  if (user != UID) {

    if (user == -1) 
      System.out.println("T " + UID  + " ServerProtocol FINISH: application " + AID + " does not exist, operation not permited");

    else

      System.out.println("T " + UID  + " ServerProtocol FINISH: User " + UID + " trying to finish application of user " + user + ", operation not permited");

    return;

  }
     

  finish(AID);
  
  
  } //end of operations on system state data

  d.getApplications().list();
  d.getSessions().list();   
  
}




 public void handleEVENT(Message a) {

  EventMessage e = (EventMessage) a;
  byte data[];
  int len;
  int AID;
  int i;


  if (!logged_in) {
       System.out.println("T " + UID  + " ServerProtocol EVENT: User not logged in, operation not permited");
       
       return;

  }
 
   // *******************  tests only ******************
 
   AID = e.getAID();
//    len = e.getLen();
//    data = e.getData();
//   System.out.println("T " + UID  + " ServerProtocol: EVENT message accepted");
//    System.out.println("T " + UID  + " ServerProtocol: Data AID" + AID + " length " + len);
//    System.out.println("T " + UID  + " ServerProtocol: Data :");
  
//    for(i = 0; i < len; i ++) {

//       System.out.print(data[i]);
//    }
//    System.out.println("");
  
   // ******************  tests only **********************

   if (AID == Const.CRLAID) {

     sendControl(a);
   }
   else {

     c.putMessage(a);

   }
}

 public void handleRJOIN(Message a) {

  RjoinMessage m = (RjoinMessage) a;
  RaskMessage rask;
  int AT = m.getAT();
  int SID = m.getSID();
  int rUID = m.getRemoteUID();
  ThreadOut rout;
  int masterUID;

  System.out.println("T " + UID  + " ServerProtocol RJOIN: RJOIN message accepted");
  System.out.println("T " + UID  + " ServerProtocol RJOIN: Data AT: " + AT + " SID: " + SID + " remoteUID: " + rUID);
  
   if (!logged_in) {
       System.out.println("T " + UID  + " ServerProtocol RJOIN: User not logged in, operation not permited");
       
       return;

    }

  synchronized(d) { //start of operations on system state data

   if (d.getSessions().getMaster(SID) == -1) {

      System.out.println("T " + UID  + " ServerProtocol RJOIN: session " + SID + " does not exist");
      return;
   }


   masterUID = d.getSessions().getMaster(SID);

   if (masterUID != UID) {

     System.out.println("T " + UID  + " ServerProtocol RJOIN: RJOIN message send by non-master user");
       
     return;

   }


   if (!d.getParticipants().isLoggedIn(rUID)) {

      System.out.println("T " + UID  + " ServerProtocol RJOIN: remote user " + rUID + " not logged in");
      return;
   }
     
  //SELECT C = count(*) FROM APPLICATIONS WHERE SID = SID AND RUID = UID
  // IF (C > 0) stop

   if (d.getApplications().inSession(SID, rUID)) {

     System.out.println("T " + UID  + " ServerProtocol RJOIN: user " + rUID + " already in session " + SID);
     return;
   }
   
   if (AT != d.getSessions().getType(SID)) { // check whether type is correct

     System.out.println("Server Protocol RJOIN: type of new application is not correct ");
     return;
   }

   //SEND_TO (RUID, RASK(AT, SID, UID(UID)))

   rask = new RaskMessage(AT, SID, UID);
   rout = c.getThread(rUID);

   rout.putMessage(rask);
  
  } //end of operations on system state data

}
  
 public void handleRANS(Message a) {

  RansMessage m = (RansMessage) a;
  int AT = m.getAT();
  int SID = m.getSID();  
  int answer = m.getAnswer();
  int rUID = m.getRemoteUID();
  int lUID = m.getLocalUID();
  int uniqueAID; 
  CommandMessage comm;
  SupdMessage update;
  String name;
  ThreadOut tmp;
  int masterUID;



  System.out.println("T " + UID  + " ServerProtocol RANS: RANS message accepted");
  System.out.println("T " + UID  + " ServerProtocol RANS: Data AT " + AT + " SID " + SID + " local UID " + lUID + " remote UID " + rUID + " answer " + answer);
 
   if (!logged_in) {
       System.out.println("T " + UID  + " ServerProtocol RANS: User not logged in, operation not permited");
       
       return;

   }

  synchronized(d) {  //start of operations on system state data

   if (d.getSessions().getMaster(SID) == -1) {

      System.out.println("T " + UID  + " ServerProtocol RANS: session " + SID + " does not exist");
      return;
   }


   if (!d.getParticipants().isLoggedIn(lUID)) {

      System.out.println("Server Protocol RANS: user " + lUID + " not logged in");
     return;
   }



   masterUID = d.getSessions().getMaster(SID);
   
   if (masterUID != lUID) {

     System.out.println("T " + UID  + " ServerProtocol RANS: RANS message send by non-master user");
       
     return;

   }

   if (rUID != UID) {

     System.out.println("T " + UID  + " ServerProtocol RANS: invalid format of RANS message, remoteUID != UID");
       
     return;
   }

  

   //IF "no" stop

   if (answer == Const.NO) {

     System.out.println("T " + UID  + " ServerProtocol RANS: user does not agree for joining session");
     tmp = c.getThread(lUID);

     if (tmp != null) {
       tmp.putMessage(a);
     }
     else {
      System.out.println("T " + UID  + " ServerProtocol RANS: user " + lUID + " not logged in");
     } 
       return;
   }

  

   if (d.getApplications().inSession(SID, UID)) {

     System.out.println("T " + UID  + " ServerProtocol RANS: user " + UID + " already in session " + SID);
     return;
   }

   if (AT != d.getSessions().getType(SID)) { // check whether type is correct

     System.out.println("T " + UID  + " ServerProtocol RANS: type of new application is not correct ");
     return;
   }


  //IF (yes) 
  //generate unique AID
    uniqueAID = assignAID();

    //INSERT INTO APPLICATIONS (AID, SID, UID(UID))
    
    d.getApplications().addApplication(uniqueAID, SID, UID);
  //  c.getSessionTable().addApplication(SID, out, uniqueAID);
    
    System.out.println("T " + UID  + " ServerProtocol RANS: application " + uniqueAID + " added to session " + SID);

      //SEND COMMAND(AT, params, AID, SID)

    comm = new CommandMessage(AT, "EMPTY", uniqueAID, SID);
      
    out.putMessage(comm);

  //  name = d.getUsers().getName(UID);

  //  update = new SupdMessage(Const.ADDUPDATE, SID, AT, UID, name);
           
      //SEND_TO_ALL SUPD(add, SID, AT, UID(UID), username)

   // sendToAll(update);
    
    } //end of operations on system state data
 
}


 public void handleMASTER(Message a) {


  MasterMessage m = (MasterMessage) a;
  int SID = m.getSID();
  MaskMessage mask;
  int master;
  ThreadOut tout;

  System.out.println("T " + UID  + " ServerProtocol MASTER: MASTER message accepted");
  System.out.println("T " + UID  + " ServerProtocol MASTER: Data SID " + SID);
 

   if (!logged_in) {
       System.out.println("T " + UID  + " ServerProtocol MASTER: User not logged in, operation not permited");
       
       return;

   }
   
   synchronized(d) {  //start of operations on system state data

   if (d.getSessions().getMaster(SID) == -1) {

      System.out.println("T " + UID  + " ServerProtocol MASTER: session " + SID + " does not exist");
      return;
   }


  //SELECT  M = MASTER FROM SESSIONS WHERE SID = SID
  master = d.getSessions().getMaster(SID);


  // if (M == UID(UID) || M == null) stop;

  if (master == UID) {
    
     System.out.println("T " + UID  + " ServerProtocol MASTER: user " + UID + " is already master of session " + SID);
     return;

  }

  if (!d.getApplications().inSession(SID, UID)) {

     System.out.println("T " + UID  + " ServerProtocol MASTER: user " + UID + " not in session " + SID);
     return;
  }
    
  tout = c.getThread(master);

    //SEND_TO(M, MASK(SID, UID(UID))

  mask = new MaskMessage(SID, UID);
  tout.putMessage(mask);
  
  
  } //end of operations on system state data

}

 public void handleMANS(Message a) {


  MansMessage m = (MansMessage) a;
  SupdMessage update;
  int answer = m.getAnswer(); 
  int SID = m.getSID();
  int lUID = m.getUID();
  Vector outVector;
  ThreadOut tmp;
  int i;
  String name;
  int master;
  int AT;

  System.out.println("T " + UID  + " ServerProtocol MANS: MANS message accepted");
  System.out.println("T " + UID  + " ServerProtocol MANS: Data SID " + SID + " UID " + lUID + " answer " + answer);
 

   if (!logged_in) {
       System.out.println("T " + UID  + " ServerProtocol MANS: User not logged in, operation not permited");
       
       return;

   }

   synchronized(d) {  //start of operations on system state data

   if (d.getSessions().getMaster(SID) == -1) {

      System.out.println("T " + UID  + " ServerProtocol MANS: session " + SID + " does not exist");
      return;
   }

   if (!d.getParticipants().isLoggedIn(lUID)) {

      System.out.println("Server Protocol MANS: user " + lUID + " not logged in");
     return;
   }


   if (!d.getApplications().inSession(SID, lUID)) {

     System.out.println("T " + UID  + " ServerProtocol MANS: user " + lUID + " not in session " + SID);
     return;
   }




   master = d.getSessions().getMaster(SID);
   
   if (master != UID) {

     System.out.println("T " + UID  + " ServerProtocol MANS: MANS message send by non-master user");
       
     return;

   }

 
   if (master == lUID) {
    
     System.out.println("T " + UID  + " ServerProtocol MANS: user " + lUID + " is already master of session " + SID);
     return;

   }


   if (!d.getParticipants().isLoggedIn(lUID)) {

      System.out.println("Server Protocol MANS: user " + lUID + " not logged in, cannot change master");
     return;
   }



   //IF "no" stop

   if (answer == Const.NO) {

     System.out.println("T " + UID  + " ServerProtocol MANS: master user does not agree for change of master");
     tmp = c.getThread(lUID);
     if (tmp != null) {
       tmp.putMessage(a);
    
     }
     else {
      System.out.println("T " + UID  + " ServerProtocol MANS: user " + lUID + " not logged in");
     } 
       return;
   }

   
   
  

   
  // if (yes) 

    //UPDATE SESSIONS SET master = UID WHERE SID = SID
    
    d.getSessions().changeMaster(SID, lUID);

    // SEND_TO_ALL (SUPD(master, SID, UID))

    name = d.getUsers().getName(lUID);
    AT = d.getSessions().getType(SID);

    update = new SupdMessage(Const.MASTERUPDATE, SID, AT, lUID, name);
      
     
      //SEND_TO_ALL SUPD(master, SID, lUID)

    sendToAll(update); 
    
    
    }  // end of operations on system state data

}

 public void handleLOGIN(Message a) {

    LoginMessage m = (LoginMessage) a;
    LogaccMessage acc;
    SupdMessage update;
    String name;
    String passwd;
    String host;
    String ip;
    DbSessions dbs;
    Applications app;
    Vector prt;
    int size;
    int i;
    int masterUID;
    int SID;
    int AT;
    int usr;
    int check;
    Integer tmp;
    String tmpname = null;
 	  long time;
	  
		time = System.currentTimeMillis();

    name = m.getUsername();
    passwd = m.getPasswd();
    host = n.getHost() + " " + (new Date(time)).toString();
    ip = n.getIP();


    System.out.println("T " + UID  + " ServerProtocol LOGIN: LOGIN message accepted");
    System.out.println("T " + UID  + " ServerProtocol LOGIN: Data username: " + name + " passwd: " + passwd);

    if (logged_in) {
       System.out.println("T " + UID  + " ServerProtocol LOGIN: User already logged in, operation not permited");
       
       return;

    }
    
	  synchronized(d) {  //start of operations on system state data
	
    //CHECK IF USER IS ALREADY LOGGED IN
       
    check = d.getUsers().getUID(name);

    if (check != -1) {

     System.out.println("T " + UID  + " ServerProtocol LOGIN: User already logged in, operation not permited"); 

     acc = new LogaccMessage(name, 0, "DENIED");

     UID = 0;

     out.putMessage(acc); 

     return;
    }
     
    // ADD new user TO USER structures
       
    d.getUsers().addUser(name, " ", " ");
    UID = d.getUsers().getUID(name);

      //SEND LOGACC with privileges = ACCESS
      
       logged_in = true;
       
       c.addThread(out, UID);  // adding threadout to table in commthread

       acc = new LogaccMessage(name, UID, d.getUsers().getPriv(UID));
       out.putMessage(acc);

       System.out.println("Server Protocol LOGIN: User logged in with UID " + UID);

      //INSERT INTO PARTICIPANTS
    
      d.getParticipants().addParticipant(UID, host, ip);

      app = d.getApplications();
      dbs = d.getSessions();
      prt = d.getParticipants().getUsers();
      
      _t.newUser(UID, this);  //add to the timeout structures

       //SEND INFORMATION ABOUT LOGGED IN USERS

      size = prt.size();

      for(i = 0; i < size; i ++) {

	tmp = (Integer) prt.elementAt(i);
	usr = tmp.intValue();
	name = d.getUsers().getName(usr) + "#" + d.getParticipants().getHost(usr) + "#" + d.getParticipants().getIP(usr);

	if (usr != UID) {
	  
	  update = new SupdMessage(Const.LOGINUPDATE, 0, 0, usr, name);
	  out.putMessage(update);

	}
      }


      // SEND_TO_ALL INFORMATION ABOUT NEW USER LOGGED IN
      
      name = d.getUsers().getName(UID) + "#" + d.getParticipants().getHost(UID) + "#" + d.getParticipants().getIP(UID);
      update =  new SupdMessage(Const.LOGINUPDATE, 0, 0, UID, name);
      
      sendToAll(update);

      // SEND INFORMATION ABOUT SESSIONS
     
      
      size = app.getSize();
      
      
      for(i = 0; i < size; i ++) {
	
	      if (app.isEnabledIndex(i)) {
		      SID = app.getSID(i);
		      usr = app.getUID(i);
		      AT = dbs.getType(SID);
			
		      tmpname = d.getUsers().getName(usr);

		      update = new SupdMessage(Const.ADDUPDATE, SID, AT, usr, tmpname);
		      out.putMessage(update);
		
	      }
      }

     
      size = dbs.getSize();

      
      //SEND INFORMATION ABOUT MASTERS

      for(i = 0; i < size; i ++) {
	
	      masterUID = dbs.getMasterUID(i);

	      if (masterUID != 0 && dbs.isEnabledIndex(i)) {
			    SID = dbs.getSID(i);
		      AT = dbs.getAT(i);
		      tmpname = d.getUsers().getName(masterUID);
		      update = new SupdMessage(Const.MASTERUPDATE, SID, AT, masterUID, tmpname);
		      out.putMessage(update);
	      }
      }
   
   } //end of operations on system state data
   
      d.getParticipants().list();
  }
 

 protected void handleLOGOUT(Message a) {

  LogoutMessage m = (LogoutMessage) a;
  int UIDpar;
  UIDpar = m.getUID();
 
 
  System.out.println("Server Protocol LOGOUT: LOGOUT message accepted");
  System.out.println("Server Protocol LOGOUT: Data UID: " + UIDpar);


  if (!logged_in) {
       System.out.println("Server Protocol LOGOUT: User not logged in, operation not permited");
       return;
  }

  //UIDpar = m.getUID();

  if (UIDpar == UID) {

    //   logout();
      
  } //IF NOT - WARNING
  else 
      System.out.println("Server Protocol LOGOUT: Wrong uid in logout message");

  synchronized(d) { //start of operations on system state data


  logout();
  
  } //end of operations on system state data

  d.getParticipants().list();
     
}

	public void handlePING(Message a) {

		_t.userResponded(UID);

	}
		
	public void handleGROUP_EVENT(Message a) {
	
		GroupEventMessage m = (GroupEventMessage) a;
		EventMessage em;
		byte data[];
		int AID;
		int newAID;
		int userlist[];
		ThreadOut tmp;
		int SID;
		int groupUID;
		int i;
		
		data = m.getData();
		AID = m.getAID();
		userlist = m.getUserList();
		SID = d.getApplications().getSession(AID);

					
	  for(i = 0; i < userlist.length; i ++) {
	 
	 	groupUID = userlist[i];		
		newAID = d.getApplications().getAID(SID, groupUID);
	  	tmp = c.getThread(groupUID);		 
		em = new EventMessage(newAID, data);

      		if (tmp != null) {
      
        	tmp.putMessage(em);
		  }		  
		}
	}
		
public void finish(int AIDpar) {

  int i;
  Vector outVector;
  ThreadOut tmp;
  SupdMessage update;
  int SID;
  boolean closed;
  int masterUID;
  int userUID;
  int AT;

  SID = d.getApplications().getSession(AIDpar);  
  userUID = d.getApplications().getUser(AIDpar);
  masterUID = d.getSessions().getMaster(SID);

  closed = d.getApplications().removeApplication(AIDpar);

  //IF (!SELECT count(*) FROM APPLICATIONS WHERE SID = s
       //DELETE FROM SESSIONS WHERE SID = s

  masterUID = d.getSessions().getMaster(SID);
  AT = d.getSessions().getType(SID); // added

  if (closed || (masterUID == userUID))  {

    System.out.println("Server Protocol finish: Session " + SID + " closed");
    d.getApplications().removeFromSession(SID);   
    d.getSessions().removeSession(SID);
    c.getSessionTable().removeSession(SID);
     //SEND_TO_ALL SUPD(remove, SID, 0)

     update = new SupdMessage(Const.REMOVEUPDATE, SID, AT, 0, "");

     }
     //stop
  
  else {
    

    System.out.println("Server Protocol finish: Application " + AIDpar + " deleted");

    //SEND_TO_ALL SUPD(remove, SID, UID(UID))
    c.getSessionTable().removeApplication(AIDpar);

    update = new SupdMessage(Const.REMOVEUPDATE, SID, AT, UID, "");

   
  }
   
  sendToAll(update);
  
  

}

public void logout() {

  Vector apps;
  Integer tmp;
  String name;
  int i;
  SupdMessage update;

 // SELECT applications FROM APPLICATIONS WHERE uid = UID

      apps = d.getApplications().getApplications(UID);

      c.removeThread(out); // from now information will not be sent to logged out user

      // FOR ALL applications FINISH

      for (i = 0; i < apps.size(); i ++) {
	
	tmp = (Integer) apps.elementAt(i);
 
	finish(tmp.intValue());
      }
      //DELETE FROM PARTICIPANTS WHERE uid = UID

			_t.userRemoved(UID);  // remove from timeout structure
			
      d.getParticipants().deleteParticipant(UID);
      
      logged_in = false;

      name = d.getUsers().getName(UID);
     
      // SEND_TO_ALL(INFORMATION ABOUT LOGGED OUT USER

      update =  new SupdMessage(Const.LOGOUTUPDATE, 0, 0, UID, name);

      sendToAll(update);

      d.getUsers().removeUser(UID);

      UID = 0;

}



private void sendToAll(Message m) {

  Vector outVector;
  ThreadOut tmp;
  int i;

  outVector = c.getAllThreads();

  for(i = 0; i < outVector.size(); i ++) {

	tmp = (ThreadOut) outVector.elementAt(i);

	tmp.putMessage(m);
	
  
  }


}


private void sendControl(Message m) {

  Vector outVector;
  ThreadOut tmp;
  int i, index;

  outVector = c.getAllThreads();
  
  index = outVector.indexOf(out);

  if (index == -1) {

    System.out.println("sendControl: cannot find sender thread");
   
  } else 

    for(i = 0; i < outVector.size(); i ++) {

      if(i != index) {

	tmp = (ThreadOut) outVector.elementAt(i);

	tmp.putMessage(m);
      
      }
  
    }

}


private static synchronized int assignAID() {

 int r;


   r = appno;
   appno ++;
   return r;
  
}

private static synchronized int assignSID() {

 int r;


   r = SIDno;
   SIDno ++;
   return r;
  
}

public void stop() {


}

}




















